/* NAME:	GReATConfigMeta
 * AUTHOR:	Attila Vizhanyo - viza@isis.vanderbilt.edu
 *
 * GReATConfig Meta Info protection & convenience proxy
 */
#ifndef GREATCONFIGMETA_H
#define GREATCONFIGMETA_H

#include <UdmBase.h>
#include <Uml.h>
#include "GR.h"

namespace GReATConfig
{
	class MetaInformation;
};

// Meta Info proxy class
// Provides a protected & convience way to access various information 
// in the Meta Info section of a GReATConfig model.
class GReATConfigMeta
{
public:
	//
	enum MetaCodeGenModes {
		MCGM_INVALID=	0x00,
		MCGM_STATIC_LIB_DEBUG=	0x10,
		MCGM_STATIC_LIB_RELEASE= 0x11,
		MCGM_DYNAMIC_LIB_DEBUG= 0x18,
		MCGM_DYNAMIC_LIB_RELEASE= 0x19,
		MCGM_NO_LIB=	0x20
	};
public:
	// ctor; user must pass a valid 'metaInfo', no validity check performed here.
	GReATConfigMeta( const GReATConfig::MetaInformation& metaInfo) :
      _metaInfo( metaInfo),
	  _sdnTransformation( GR::diagram),
      _dn( Uml::diagram)
		{}
	// dtor
	virtual ~GReATConfigMeta() 
		{}
	// Return 'path' if 'path' is absolute
	// Return 'currWD'\'path' if 'path' is relative
	static std::string getFullPath( const std::string& currWD, const std::string& path);
	//
	static bool isLibraryToGenerate( MetaCodeGenModes mcgm);
	static bool isLibraryDebug( MetaCodeGenModes mcgm);
	static bool isLibraryRelease( MetaCodeGenModes mcgm);
	static bool isLibraryStatic( MetaCodeGenModes mcgm);
	static bool isLibraryDynamic( MetaCodeGenModes mcgm);

	// Opens(!) the transformation dn specified in '_metaInfo' and returns the root folder of the dn.
	virtual void getRootFolder( GR::RootFolder& rootFolder, const std::string& cfgFileDir= "");
	// Returns GR Transformation data network. 
	Udm::DataNetwork& getTransformationDN( const std::string& cfgFileDir= "");
	// Returns with a map of [Diagram name of dn,dn pointer] specified in '_metaInfo'
	// If udm project file was specified, returns all participating dns but not the cross link dn. 
	virtual void getGeneralBackend( std::map< std::string, const Udm::DataNetwork*>& dns, const std::string& cfgFileDir= "");
	// Returns true, if udm project file was specified in '_metaInfo'.
	bool isUdmProjectFile();
	// Returns the name of the udm project file.
	std::string getUdmProjectFileName();
	// Returns the name of the cross link diagram. 
	// If not udm project was specified, returns empty string.
	std::string getCrossLinkDiagramName();
	// Returns the cross link diagram of the cross link dn in 'xLinkDiagram'.
	// If not udm project was specified, returns empty string.
	void getCrossLinkDiagram( Uml::Diagram& xLinkDiagram);
	// Returns reference to the cross link dn.
	// If not udm project was specified, exception is thrown by udm (udm_exception)
	Udm::DataNetwork& getCrossLinkDN();
	// Returns with the name of the transformation file specified in '_metaInfo'.
	std::string getTransformationFileName();
	//
	std::string incFiles() const;
	std::string libFiles() const;
	std::string incDirs() const;
	std::string libDirs() const;
	//
	MetaCodeGenModes getMetaCodeGenMode() const;

private:
	// Reference to meta info section of GReATConfig model.
	const GReATConfig::MetaInformation& _metaInfo;
	// Data network for the transformation file.
	Udm::SmartDataNetwork _sdnTransformation;
	// Meta information: udm project OR data network.
	Udm::SmartDataNetwork _dn;
	Udm::UdmProject _prj;
};

// tokenize comma-separated-value lines
extern int tokenize_csv_lines( const std::string& line, std::vector< std::string>& tokens);
 
#endif // GREATCONFIGMETA_H
