/* NAME:	Input filter
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Generate code, which performs identity check filtering. 
 */
#ifndef __INPUTFILTER_H
#define __INPUTFILTER_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include <map>
#include "RuleTypes.h"
#include "PE.h"
#include "PacketPassing.h"

// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
// 
// USE it either by calling execute(), which generate code for all inputs, -OR-
// call createFunction(), createInitCode() only once, then call 
// createUniqueInputMember(), inputIdentityCheck() for each input, finally 
// call createFinalCode() only once. (one-by-one parallel processing).
class InputFilter
{
public:
	// ctor; see parameter desc. at the member variables declaration part. 
	InputFilter( const GR::RewritingRule& rule,
		const std::map< GR::ObjectWrapper, std::string>* boundObjects = 0
		) : _rule( rule) , _boundObjects( boundObjects)
		{}
	// Create input identity check filtering code for each input in 'ruleInputs' into 'addTo' and 'addCallTo'.
	void execute( ParamDatas_t& ruleInputs, Class& addTo, PEComposite& addCallTo);
	// Create input filter function with name 'isInputUnique' into 'filterFuncDef' and a call into 'filterFuncCall'
	// If 'addDefTo' and/ or 'addCallTo' are not 0, they are appended with these created PEs.
	void createFunction( Function** filterFuncDef, FunctionCall** filterFuncCall, Class* addDefTo= 0, PEComposite* addCallTo= 0);
	// Create initalizing code into 'filterFuncDef'. 
	// Returns with the 'uniqueCondition' PE, which is later used in createFinalCode() and inputIdentityCheck().
	void createInitCode( Function* filterFuncDef, Loop** inputLoop, If** uniqueCondition, If** checkForUnique);
	// Create finalizing code into 'filterFuncDef'. Caller pass the return value of createInitCode() as 'uniqueCondition'.
	void createFinalCode( If* uniqueCondition, Function* filterFuncDef);
	// Create member variable for each input in 'ruleInputs' into 'addTo'.
	void createUniqueInputMembers( ParamDatas_t& ruleInputs, Class& addTo);
	// Create member variable for 'ruleInput' into 'addTo'.
	void createUniqueInputMember( ParamData& ruleInput, Class& addTo);
	// Create input identity check filtering code for 'ruleInput' into 'filterFuncDef', 'filterFuncCall', 
	// 'uniqueCondition' (pass the return value of createInitCode()) and currHitCheck (pass the return value of inputIdentityCheck())
	// 'callerParamName' is the name of input variable passed in 'filterFuncCall'.
	void inputIdentityCheck( 
			const ParamData& ruleInput, 
			const std::string& callerParamName, 
			Function* filterFuncDef, 
			FunctionCall* filterFuncCall,
			Loop* inputLoop,
			If* uniqueCondition,
			If* checkForUnique
			);
	// Create input identity check filtering code for each input in 'ruleInputs' into 'filterFuncDef' and 'filterFuncCall'. 
	// Same as createInitCode(), inputIdentityCheck() for each input, createFinalCode().
	void inputIdentityChecks( const ParamDatas_t& ruleInputs, Function* filterFuncDef, FunctionCall* filterFuncCall);

protected:
	// Return with generated function parameter 'param' for 'input'.
	void createParam( const GR::ObjectWrapper& input, FuncParam_t& param) const;

private:
	// Rule, on which output filter code generation executes.
	const GR::RewritingRule& _rule;
	// Maps object wrappers to their generated variable names. 
	const std::map< GR::ObjectWrapper, std::string>* _boundObjects;
};

#endif //__INPUTFILTER_H