/* NAME:	PE - Program Element
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Program Elements (PEs) represent C++ code building blocks
 * like loops, functions, etc..
 * Object hierarchy follows the Composite Design Pattern.
 */
#ifndef __PE_H
#define __PE_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include <list>
#include <string>
#include "CGEx.h"

// Program Element ancestor class for all PEs (abstract).
class PE
{
public:
	virtual ~PE() {}
	// Prints the PE to 'os'. 
	// 'operator<<()' cannot be virtual, so 'operator<<()' calls 'print()'.
	virtual std::ostream& print( std::ostream& os) const = 0;
	// Prints the PE to header file. Default implementation does nothing.
	virtual std::ostream& printHeader( std::ostream& os) const 
	{
		return os;	// do nothing
	}
	// Prints the PE to source file. Default implementation calls 'print()' on itself.
	virtual std::ostream& printSource( std::ostream& os) const
	{
		return print( os);
	}
	// Composite pattern functions. They raise an exception by default, see composite descendants override it.
	virtual void add( const PE&)
	{ 
		throw CGEx("Trying to add PE to a non-composite instance.");
	}
	virtual void remove( const PE&)
	{
		throw CGEx("Trying to delete PE from a non-composite instance.");
	}
	// Support unique PE identifiers.
	static void resetPE_UID();
	static std::string getPE_UID();
	static std::string createUniqueName( const std::string& name, bool decorateName= true);

protected:
	// Current tabulation position.
	static unsigned short _tab;
	static std::ostream& printTab( std::ostream& os);

private:
	// Program wide unique PE id.
	static unsigned long _id;
};

// PE output stream, virtuality is achieved through 'print()'.
std::ostream& operator<<( std::ostream& os, const PE& pe);

// Composite base class, for managing composite structures.
class PEComposite : public PE
{
public:
	// dtor. A 'PEComposite' type object is ALWAYS responsible for deleting its children.
	virtual ~PEComposite()
	{
		for( PEs_t::iterator itElems= _elems.begin(); itElems != _elems.end(); ++itElems)
			delete (*itElems);
	}
	// Printers call the proper printer functions on '_elems'.
	virtual std::ostream& print( std::ostream &os) const;
	virtual std::ostream& printHeader( std::ostream &os) const;
	virtual std::ostream& printSource( std::ostream &os) const;
	// Composite pattern functions
	virtual void add( const PE& elem) 
	{
		_elems.push_back( &elem);
	}
	virtual void remove( const PE& elem) 
	{
		_elems.remove( &elem);
	}
	virtual int size() const
	{
		return _elems.size();
	}

private:
	friend class Class;	// Class needs to be able to access '_elems'.
	typedef std::list< const PE*> PEs_t;
	PEs_t _elems;
};

// Function parameter. 'pair' must be wrapped into a different type, to be able to compare
// parameters upon there types regardless of there names. See: comparision operator below. 
struct FuncParam_t 
{
	FuncParam_t() {}
	FuncParam_t( const std::pair< std::string, std::string>& param) : _param( param) {}
	std::pair< std::string, std::string> _param; // type, variable
};	

// Comparision operator of Function params. Two params are considered to be equal, if there types are equal.
bool operator==( const FuncParam_t& cmp1, const FuncParam_t& cmp2);

// Function parameters.
typedef std::list< FuncParam_t> FuncParams_t;

// C function definition
class Function : public PEComposite
{
public:
	// Prints signature and body.
	virtual std::ostream& print( std::ostream &os) const;
	// Prints signature only.
	virtual std::ostream& printHeader( std::ostream &os) const;
	//
	virtual std::ostream& printSource( std::ostream& os) const
	{
		return print( os);
	}
	// Parameter creators.
	static FuncParam_t createParam( const std::string& namesp, const std::string& type, const std::string& name, bool decorateName= true);
	static FuncParam_t createInputParam( const std::string& name, bool decorateName= true);
	static FuncParam_t createOutputParam( const std::string& name, bool decorateName= true);
	// Input parameters are inserted after the last input parameter, but before the first output parameter.
	const FuncParam_t& insertInputParam( const FuncParam_t& inputParam);
	// Output parameters are inserted after output parameters. (Thus, at the end of the parameter list.)
	const FuncParam_t& insertOutputParam( const FuncParam_t& outputParam);

protected:
	std::ostream& printSignature( std::ostream &os, bool printScope) const;
	std::ostream& printBody( std::ostream &os) const;

public:
	std::string _scope;
	std::string _name;
	std::string _returnValue;
	FuncParams_t _params;
};


// C for loop
class Loop : public PEComposite
{
public:
	Loop() : PEComposite(), _isFirstIterator( true) {}
	virtual std::ostream& print( std::ostream &os) const;
	// add iterator variable to loop. 
	// (only same iterators with same 'containerType'-s should be added, however, that is not checked.)
	void addIterator( const std::string& iter, const std::string& container, const std::string& containerType);
	// add iterator variable to loop with type 'PKS'.
	void addInputIterator( const std::string& iter, const std::string& container);
	// create a loop with one iterator: variable 'iterator', of type 'containerType', on container 'container'
	void createIteratorLoop( const std::string& iter, const std::string& container, const std::string& containerType);
	// create a loop: for( std::list< Match>::const_iterator itMatch= _matches.begin(); itMatch!= _matches.end(); ++itMatch)
	// and add to its body the definition: Match currMatch= *itMatch;
	void createMatchLoop();

public:
	std::string _loopVariable;
	std::string _from;
	std::string _to;
	std::string _increment;
	// by adding iterators to loop header, the first and subsequent iterators are generated differently.
	bool _isFirstIterator;
};

// C if condition
class If : public PEComposite
{
public:
	virtual std::ostream& print( std::ostream &os) const;
	// Adding elements to the optional else branch.
	void addElse( const PE& elem) 
	{
		_elseBranch.add( elem);
	}
	// Removing elements from the optional else branch.
	void removeElse( const PE& elem) 
	{
		_elseBranch.remove( elem);
	}
	// Size of the optional else branch.
	int sizeElse() const
	{
		return _elseBranch.size();
	}
	// Returns with reference to the else branch. (This function should be used rarely.)
	PEComposite& getElseBranch()
	{ 
		return _elseBranch;
	}
	// Append '_expression' with 'expressionTag' with logical AND semantics.
	void appendWithAnd( const std::string& expressionTag);
	// Append '_expression' with 'expressionTag' with logical OR semantics.
	void appendWithOr( const std::string& expressionTag);
	// Return true, if '_expression' is not empty.
	bool isExpression() const;
	// Creates the type check condition: "false== Uml::IsDerivedFrom( 'varName'.type(), 'varParadigm'::'varType'::meta)", 
	// and adds a 'continue' expression as a body.
	void createIfTypeMismatchThenContinue( const std::string& varName, const std::string& varParadigm, const std::string& varType);
	// Creates the condition: "expr" and adds a 'return' expression as a body.
	void createIfExprThenReturn( const std::string& expr);
	// Creates the condition: "expr" and adds a 'continue' expression as a body.
	void createIfExprThenContinue( const std::string& expr);
	// Creates the conditon: "<outputMemberName>->end()!= <findName>", and adds "isUnique= true" as a body.
	void createHitCheckCondition( const std::string& outputMemberName, const std::string& findName);

public:
	std::string _expression;

private:
	// Prints 'branch' to 'os'.
	std::ostream& printBranch( std::ostream& os, const PEComposite& branch) const;
	// Else branch.
	PEComposite _elseBranch;
};

// C case 
// !!VIZA currently not used.
class Case : public PEComposite
{
public:
	Case() : PEComposite() , _break( true) {}
	virtual std::ostream& print( std::ostream &os) const;

public:
	std::string _value;
	bool _break;
};

// C switch
// !!VIZA currently not used.
class Switch  : public PEComposite
{
public:
	virtual std::ostream& print( std::ostream &os) const;

public:
	std::string _switchVariable;
};

// Class data member 
class ClassDef : public PE
{
public:
	//
	virtual std::ostream& print( std::ostream &os) const;
	// Class definitions are printed only into header files.
	virtual std::ostream& printHeader( std::ostream &os) const
	{
		return print( os);
	}
	// Class definitions are not printed into source files.
	virtual std::ostream& printSource( std::ostream &os) const
	{
		return os; // do nothing
	}
	// Creates "PKS <name>;".
	void createPackets( const std::string& name);
	// Creates "PKS* <name>;".
	void createPacketsPtr( const std::string& name);

public:
	std::string _type;
	std::string _variable;
};

// C declaration/ definition
class Definition : public ClassDef
{
public:
	virtual std::ostream& print( std::ostream &os) const;
	// Add an argument to '_args'. Arguments are for '_variable'.
	void addArg( const std::string& arg) { _args.push_back( arg);}
	// Create the common UDM program line: "set< 'paradigm'::'type'> 'variable'= 'parent'.'type'_kind_children();"
	void createContainmentDef( const std::string& paradigm, const std::string& type, const std::string& variable, const std::string& parent);
	// Create the definition: "const Udm::Object& <currInputVar>= *<iter>;"
	void createCurrentInput( const std::string& currInputVar, const std::string& iter);
	// Create the commmon UDM program line:"set< 'paradigm'::'type'> 'variable'= 'parent'.'role'();"
	void createRoles( const std::string& paradigm, const std::string& type, const std::string& variable, const std::string& parent, const std::string& role);
	// Create the commmon UDM program line:"'paradigm'::'type' 'variable'= 'parent'.'role'();"
	void createRole( const std::string& paradigm, const std::string& type, const std::string& variable, const std::string& parent, const std::string& role);
	// Create the definition: "<type> <variable>= <type>::Cast( <varToCast>);"
	void createTypeCast( const std::string& type, const std::string& variable, const std::string& varToCast);
	// Create the definition: "Udm::Object::AssociationInfo <variable>( Uml::Class( NULL));"
	void createAssocInfo( const std::string& variable);
	// Create the definition: "<paradigm>::<type> <variable>= <paradigm>::<type>::Create( <parent> [,<childRole>]);"
	void createUdmCreate( const std::string& paradigm, const std::string& type, const std::string& variable, const std::string& parent, const std::string& childRole);

public:
	std::string _initValue;

private:
	// Constructor parameters of class '_type'.
	std::list< std::string> _args;
};

// C expression
class Expression : public PE
{
public:
	Expression( const std::string& expression= "") : PE(), _expression( expression) {}
	virtual std::ostream& print( std::ostream &os) const;

public:
	std::string _expression;
};

// C++ style comment
class CommentLine : public PE
{
public:
	CommentLine() : PE() {}
	CommentLine( const std::string& comment) : PE(), _comment( comment) {}
	virtual std::ostream& print( std::ostream& os) const;

public:
	std::string _comment;
};

// C macro #ifdef
class Macro : public PEComposite
{
	virtual std::ostream& print( std::ostream &os) const;

public:
	std::string _defVar;
};

// C++ class
class Class : public PE
{
public:
	// Prints everything into one file. (Functions definitions are inside class declaration}
	virtual std::ostream& print( std::ostream &os) const;
	// Prints class header file (Without function definitions)
	virtual std::ostream& printHeader( std::ostream& os) const;
	// Prints class source file (Only function definitions)
	virtual std::ostream& printSource( std::ostream& os) const;
	// Prints forward declaration
	virtual std::ostream& printFwdDecl( std::ostream& os) const;
	// Add 'PE'-s to 'public', 'protected', 'private' blocks of class definition.
	void addPublic( const PE& elem) { _public.add( elem); }
	void addProtected( const PE& elem) { _protected.add( elem); }
	void addPrivate( const PE& elem) { _private.add( elem); }
	// Addition of functions to classes involves setting the function scope to the name of the class.
	void addPublic( Function& funcDef) { funcDef._scope= _name; _public.add( funcDef); }
	void addProtected( Function& funcDef) { funcDef._scope= _name; _protected.add( funcDef); }
	void addPrivate( Function& funcDef) { funcDef._scope= _name; _private.add( funcDef); }
	// Gain access to 'PE'-s added to 'public', 'protected', 'private' blocks of class definition.
	const PEComposite::PEs_t* getPublic() const { return &_public._elems;}
	const PEComposite::PEs_t* getProtected() const { return &_protected._elems;}
	const PEComposite::PEs_t* getPrivate() const { return &_private._elems;}

public:
	std::string _name;

private:
	PEComposite _public;
	PEComposite _protected;
	PEComposite _private;
};

// Caller paremeters.
typedef std::list< std::string> CallParams_t;

// Abstract base for class FunctionCall and class FuncOperator.
class Call : public PE
{
public:
	std::ostream& printCall( const std::string& name, std::ostream& os) const;

public:
	CallParams_t _params;
	std::string _retType;
	std::string _retVar;
};

// C function call. Cannot exist without a matching function definition (class Function).
class FunctionCall : public Call
{
public:
	FunctionCall() : _func( NULL) {}
	virtual std::ostream& print( std::ostream &os) const;
	// copy the parameter names from the pointed function definition.
	void copyParamNames();

public:
	Function* _func;
};

class FuncOperator : public Call
{
public:
	FuncOperator() : _class( NULL) {}
	virtual std::ostream& print( std::ostream &os) const;

public:
	const Class* _class;
	std::string _name;	// ...of instantiated object. If empty, then name will be lowerBegin( _class->_name).
};

class TryCatchBlock : public PEComposite
{
public:
	virtual std::ostream& print( std::ostream &os) const;
	// Adding elements to the optional catch block.
	virtual void addCatch( const PE& elem) 
	{
		_catchBlock.add( elem);
	}
	// Removing elements from the optional catch block.
	virtual void removeCatch( const PE& elem) 
	{
		_catchBlock.remove( elem);
	}
	// Size of the optional catch block.
	virtual int sizeCatch() const
	{
		return _catchBlock.size();
	}

public:
	std::string _catchExpr;

private:
	// Prints 'block' to 'os'.
	std::ostream& printBlock( std::ostream& os, const PEComposite& block) const;
	// Catch block composite.
	PEComposite _catchBlock;
};

// C++ program. Topmost element of program hierarchy.
class Program : public PEComposite
{
public:
	virtual std::ostream& print( std::ostream &os) const;
	virtual std::ostream& printSource( std::ostream &os) const;
	virtual std::ostream& printHeader( std::ostream& os) const;
	// program consist of classes only!
	virtual void add( const Class& cl) 
	{
		_classes.push_back( &cl);
	}
	virtual void remove( const Class& cl) 
	{
		_classes.remove( &cl);
	}
	virtual int size() const
	{
		return _classes.size();
	}

private:
	// Class list.
	typedef std::list< const Class*> Classes_t;
	Classes_t _classes;
};

#endif //__PE_H